@@ -2,8 +2,8 @@
  *
  *  BlueZ - Bluetooth protocol stack for Linux
  *
- *  Copyright (C) 2006-2007  Nokia Corporation
- *  Copyright (C) 2004-2008  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2006-2010  Nokia Corporation
+ *  Copyright (C) 2004-2010  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -31,8 +31,6 @@
 #include <string.h>
 
 #include <bluetooth/bluetooth.h>
-#include <bluetooth/hci.h>
-#include <bluetooth/hci_lib.h>
 #include <bluetooth/sdp.h>
 #include <bluetooth/sdp_lib.h>
 
@@ -43,11 +41,11 @@
 #include "plugin.h"
 #include "adapter.h"
 #include "error.h"
-#include "logging.h"
+#include "log.h"
 
 #define SERVICE_INTERFACE "org.bluez.Service"
 
-static DBusConnection *connection = NULL;
+static DBusConnection *connection;
 
 struct record_data {
 	uint32_t handle;
@@ -77,6 +75,8 @@ struct service_adapter {
 	GSList *records;
 };
 
+static struct service_adapter *serv_adapter_any = NULL;
+
 static int compute_seq_size(sdp_data_t *data)
 {
 	int unit_size = data->unitSize;
@@ -105,7 +105,7 @@ static void element_start(GMarkupParseContext *context,
 				break;
 			}
 		}
-		debug("New attribute 0x%04x", ctx_data->attr_id);
+		DBG("New attribute 0x%04x", ctx_data->attr_id);
 		return;
 	}
 
@@ -172,13 +172,13 @@ static void element_end(GMarkupParseContext *context,
 			int ret = sdp_attr_add(ctx_data->record, ctx_data->attr_id,
 							ctx_data->stack_head->data);
 			if (ret == -1)
-				debug("Trouble adding attribute\n");
+				DBG("Trouble adding attribute\n");
 
 			ctx_data->stack_head->data = NULL;
 			sdp_xml_data_free(ctx_data->stack_head);
 			ctx_data->stack_head = NULL;
 		} else {
-			debug("No data for attribute 0x%04x\n", ctx_data->attr_id);
+			DBG("No data for attribute 0x%04x\n", ctx_data->attr_id);
 		}
 		return;
 	}
@@ -319,7 +319,7 @@ static void exit_callback(DBusConnection *conn, void *user_data)
 	struct service_adapter *serv_adapter = user_record->serv_adapter;
 	struct pending_auth *auth;
 
-	debug("remove record");
+	DBG("remove record");
 
 	serv_adapter->records = g_slist_remove(serv_adapter->records,
 						user_record);
@@ -357,7 +357,7 @@ static inline DBusMessage *failed(DBusMessage *msg)
 static inline DBusMessage *failed_strerror(DBusMessage *msg, int err)
 {
 	return g_dbus_create_error(msg, ERROR_INTERFACE ".Failed",
-			strerror(err));
+							"%s", strerror(err));
 }
 
 static inline DBusMessage *not_authorized(DBusMessage *msg)
@@ -386,7 +386,11 @@ static int add_xml_record(DBusConnection *conn, const char *sender,
 		return -EIO;
 	}
 
-	adapter_get_address(serv_adapter->adapter, &src);
+	if (serv_adapter->adapter)
+		adapter_get_address(serv_adapter->adapter, &src);
+	else
+		bacpy(&src, BDADDR_ANY);
+
 	if (add_record_to_server(&src, sdp_record) < 0) {
 		error("Failed to register service record");
 		sdp_record_free(sdp_record);
@@ -401,9 +405,9 @@ static int add_xml_record(DBusConnection *conn, const char *sender,
 					exit_callback, user_record, NULL);
 
 	serv_adapter->records = g_slist_append(serv_adapter->records,
-					user_record);
+								user_record);
 
-	debug("listener_id %d", user_record->listener_id);
+	DBG("listener_id %d", user_record->listener_id);
 
 	*handle = user_record->handle;
 
@@ -424,7 +428,10 @@ static DBusMessage *update_record(DBusConnection *conn, DBusMessage *msg,
 				"Not Available");
 	}
 
-	adapter_get_address(serv_adapter->adapter, &src);
+	if (serv_adapter->adapter)
+		adapter_get_address(serv_adapter->adapter, &src);
+	else
+		bacpy(&src, BDADDR_ANY);
 
 	sdp_record->handle = handle;
 	err = add_record_to_server(&src, sdp_record);
@@ -433,7 +440,7 @@ static DBusMessage *update_record(DBusConnection *conn, DBusMessage *msg,
 		error("Failed to update the service record");
 		return g_dbus_create_error(msg,
 				ERROR_INTERFACE ".Failed",
-				strerror(EIO));
+				"%s", strerror(EIO));
 	}
 
 	return dbus_message_new_method_return(msg);
@@ -472,7 +479,7 @@ static DBusMessage *update_xml_record(DBusConnection *conn,
 		sdp_record_free(sdp_record);
 		return g_dbus_create_error(msg,
 				ERROR_INTERFACE ".Failed",
-				strerror(EIO));
+				"%s", strerror(EIO));
 	}
 
 	return update_record(conn, msg, serv_adapter, handle, sdp_record);
@@ -484,13 +491,13 @@ static int remove_record(DBusConnection *conn, const char *sender,
 {
 	struct record_data *user_record;
 
-	debug("remove record 0x%x", handle);
+	DBG("remove record 0x%x", handle);
 
 	user_record = find_record(serv_adapter, handle, sender);
 	if (!user_record)
 		return -1;
 
-	debug("listner_id %d", user_record->listener_id);
+	DBG("listner_id %d", user_record->listener_id);
 
 	g_dbus_remove_watch(conn, user_record->listener_id);
 
@@ -568,7 +575,6 @@ static void auth_cb(DBusError *derr, void *user_data)
 	}
 
 	if (derr) {
-		adapter_get_address(serv_adapter->adapter, &src);
 		error("Access denied: %s", derr->message);
 
 		reply = not_authorized(auth->msg);
@@ -584,16 +590,20 @@ done:
 	dbus_connection_unref(auth->conn);
 
 	serv_adapter->pending_list = g_slist_remove(serv_adapter->pending_list,
-						auth);
+									auth);
 	g_free(auth);
 
 	auth = next_pending(serv_adapter);
 	if (auth == NULL)
 		return;
 
-	adapter_get_address(serv_adapter->adapter, &src);
+	if (serv_adapter->adapter)
+		adapter_get_address(serv_adapter->adapter, &src);
+	else
+		bacpy(&src, BDADDR_ANY);
+
 	btd_request_authorization(&src, &auth->dst,
-				auth->uuid, auth_cb, serv_adapter);
+					auth->uuid, auth_cb, serv_adapter);
 }
 
 static DBusMessage *request_authorization(DBusConnection *conn,
@@ -601,7 +611,6 @@ static DBusMessage *request_authorization(DBusConnection *conn,
 {
 	struct record_data *user_record;
 	struct service_adapter *serv_adapter = data;
-	struct btd_adapter *adapter = serv_adapter->adapter;
 	sdp_record_t *record;
 	sdp_list_t *services;
 	const char *sender;
@@ -622,10 +631,15 @@ static DBusMessage *request_authorization(DBusConnection *conn,
 		return failed(msg);
 
 	user_record = find_record(serv_adapter, handle, sender);
-	if (!user_record)
-		return not_authorized(msg);
+	if (!user_record) {
+		user_record = find_record(serv_adapter_any, handle, sender);
+		if (!user_record)
+			return not_authorized(msg);
+	}
 
 	record = sdp_record_find(user_record->handle);
+	if (record == NULL)
+		return not_authorized(msg);
 
 	if (sdp_get_service_classes(record, &services) < 0) {
 		sdp_record_free(record);
@@ -654,16 +668,21 @@ static DBusMessage *request_authorization(DBusConnection *conn,
 	str2ba(address, &auth->dst);
 
 	serv_adapter->pending_list = g_slist_append(serv_adapter->pending_list,
-			auth);
+									auth);
 
 	auth = next_pending(serv_adapter);
 	if (auth == NULL)
 		return does_not_exist(msg);
 
-	adapter_get_address(adapter, &src);
+	if (serv_adapter->adapter)
+		adapter_get_address(serv_adapter->adapter, &src);
+	else
+		bacpy(&src, BDADDR_ANY);
+
 	if (btd_request_authorization(&src, &auth->dst, auth->uuid, auth_cb,
-					serv_adapter) < 0) {
-		serv_adapter->pending_list = g_slist_remove(serv_adapter->pending_list, auth);
+							serv_adapter) < 0) {
+		serv_adapter->pending_list = g_slist_remove(serv_adapter->pending_list,
+									auth);
 		g_free(auth);
 		return not_authorized(msg);
 	}
@@ -676,7 +695,6 @@ static DBusMessage *cancel_authorization(DBusConnection *conn,
 {
 	DBusMessage *reply;
 	struct service_adapter *serv_adapter = data;
-	struct btd_adapter *adapter = serv_adapter->adapter;
 	struct pending_auth *auth;
 	const gchar *sender;
 	bdaddr_t src;
@@ -687,7 +705,11 @@ static DBusMessage *cancel_authorization(DBusConnection *conn,
 	if (auth == NULL)
 		return does_not_exist(msg);
 
-	adapter_get_address(adapter, &src);
+	if (serv_adapter->adapter)
+		adapter_get_address(serv_adapter->adapter, &src);
+	else
+		bacpy(&src, BDADDR_ANY);
+
 	btd_cancel_authorization(&src, &auth->dst);
 
 	reply = not_authorized(auth->msg);
@@ -697,16 +719,20 @@ static DBusMessage *cancel_authorization(DBusConnection *conn,
 	dbus_connection_unref(auth->conn);
 
 	serv_adapter->pending_list = g_slist_remove(serv_adapter->pending_list,
-						auth);
+									auth);
 	g_free(auth);
 
 	auth = next_pending(serv_adapter);
 	if (auth == NULL)
 		goto done;
 
-	adapter_get_address(adapter, &src);
+	if (serv_adapter->adapter)
+		adapter_get_address(serv_adapter->adapter, &src);
+	else
+		bacpy(&src, BDADDR_ANY);
+
 	btd_request_authorization(&src, &auth->dst,
-				auth->uuid, auth_cb, serv_adapter);
+					auth->uuid, auth_cb, serv_adapter);
 
 done:
 	return dbus_message_new_method_return(msg);
@@ -725,42 +751,67 @@ static GDBusMethodTable service_methods[] = {
 static void path_unregister(void *data)
 {
 	struct service_adapter *serv_adapter = data;
+	GSList *l, *next = NULL;
 
-	g_slist_foreach(serv_adapter->records, (GFunc) exit_callback, NULL);
+	for (l = serv_adapter->records; l != NULL; l = next) {
+		struct record_data *user_record = l->data;
+
+		next = l->next;
+
+		g_dbus_remove_watch(connection, user_record->listener_id);
+		exit_callback(connection, user_record);
+	}
+
+	g_free(serv_adapter);
 }
 
-static int service_probe(struct btd_adapter *adapter)
+static int register_interface(const char *path, struct btd_adapter *adapter)
 {
-	const char *path = adapter_get_path(adapter);
 	struct service_adapter *serv_adapter;
 
 	DBG("path %s", path);
 
-	serv_adapter = g_new0(struct service_adapter, 1);
+	serv_adapter = g_try_new0(struct service_adapter, 1);
+	if (serv_adapter == NULL)
+		return -ENOMEM;
+
 	serv_adapter->adapter = adapter;
 	serv_adapter->pending_list = NULL;
 
-	if (!g_dbus_register_interface(connection, path,
-					SERVICE_INTERFACE,
-					service_methods, NULL, NULL,
-					serv_adapter, path_unregister)) {
+	if (g_dbus_register_interface(connection, path, SERVICE_INTERFACE,
+				service_methods, NULL, NULL, serv_adapter,
+						path_unregister) == FALSE) {
 		error("D-Bus failed to register %s interface",
-				SERVICE_INTERFACE);
-		return -1;
+							SERVICE_INTERFACE);
+		g_free(serv_adapter);
+		return -EIO;
 	}
 
-	info("Registered interface %s on path %s", SERVICE_INTERFACE, path);
+	DBG("Registered interface %s on path %s", SERVICE_INTERFACE, path);
+
+	if (serv_adapter->adapter == NULL)
+		serv_adapter_any = serv_adapter;
+
+	return 0;
+}
+
+static void unregister_interface(const char *path)
+{
+	DBG("path %s", path);
+
+	g_dbus_unregister_interface(connection, path, SERVICE_INTERFACE);
+}
+
+static int service_probe(struct btd_adapter *adapter)
+{
+	register_interface(adapter_get_path(adapter), adapter);
 
 	return 0;
 }
 
 static void service_remove(struct btd_adapter *adapter)
 {
-	const char *path = adapter_get_path(adapter);
-
-	DBG("path %s", path);
-
-	g_dbus_unregister_interface(connection, path, SERVICE_INTERFACE);
+	unregister_interface(adapter_get_path(adapter));
 }
 
 static struct btd_adapter_driver service_driver = {
@@ -769,20 +820,46 @@ static struct btd_adapter_driver service_driver = {
 	.remove	= service_remove,
 };
 
+static const char *any_path;
+
 static int service_init(void)
 {
+	int err;
+
 	connection = dbus_bus_get(DBUS_BUS_SYSTEM, NULL);
 	if (connection == NULL)
 		return -EIO;
 
-	return btd_register_adapter_driver(&service_driver);
+	any_path = btd_adapter_any_request_path();
+	if (any_path != NULL) {
+		if (register_interface(any_path, NULL) < 0) {
+			btd_adapter_any_release_path();
+			any_path = NULL;
+		}
+	}
+
+	err = btd_register_adapter_driver(&service_driver);
+	if (err < 0) {
+		dbus_connection_unref(connection);
+		return err;
+	}
+
+	return 0;
 }
 
 static void service_exit(void)
 {
 	btd_unregister_adapter_driver(&service_driver);
 
+	if (any_path != NULL) {
+		unregister_interface(any_path);
+
+		btd_adapter_any_release_path();
+		any_path = NULL;
+	}
+
 	dbus_connection_unref(connection);
 }
 
-BLUETOOTH_PLUGIN_DEFINE("service", service_init, service_exit)
+BLUETOOTH_PLUGIN_DEFINE(service, VERSION,
+		BLUETOOTH_PLUGIN_PRIORITY_HIGH, service_init, service_exit)
