@@ -34,8 +34,6 @@
 #include <sys/ioctl.h>
 
 #include <bluetooth/bluetooth.h>
-#include <bluetooth/hci.h>
-#include <bluetooth/hci_lib.h>
 #include <bluetooth/sdp.h>
 #include <bluetooth/sdp_lib.h>
 
@@ -202,6 +200,9 @@ static void dev_info_free(struct remote_dev_info *dev)
 {
 	g_free(dev->name);
 	g_free(dev->alias);
+	g_slist_foreach(dev->services, (GFunc) g_free, NULL);
+	g_slist_free(dev->services);
+	g_strfreev(dev->uuids);
 	g_free(dev);
 }
 
@@ -275,8 +276,7 @@ static void update_ext_inquiry_response(struct btd_adapter *adapter)
 	memset(data, 0, sizeof(data));
 
 	if (dev->ssp_mode > 0)
-		create_ext_inquiry_response((char *) dev->name,
-						adapter->tx_power,
+		create_ext_inquiry_response(dev->name, adapter->tx_power,
 						adapter->services, data);
 
 	ret = adapter_ops->write_eir_data(adapter->dev_id, data);
@@ -309,8 +309,8 @@ int adapter_set_service_classes(struct btd_adapter *adapter, uint8_t value)
 
 	err = adapter_ops->set_class(adapter->dev_id, adapter->wanted_cod);
 	if (err < 0)
-		error("Adapter class update failed: %s(%d)",
-						strerror(err), err);
+		error("Adapter class update failed: %s (%d)",
+						strerror(-err), -err);
 	else
 		adapter->pending_cod = adapter->wanted_cod;
 
@@ -335,8 +335,8 @@ int btd_adapter_set_class(struct btd_adapter *adapter, uint8_t major,
 
 	err = adapter_ops->set_class(adapter->dev_id, adapter->wanted_cod);
 	if (err < 0)
-		error("Adapter class update failed: %s(%d)",
-						strerror(err), err);
+		error("Adapter class update failed: %s (%d)",
+						strerror(-err), -err);
 	else
 		adapter->pending_cod = adapter->wanted_cod;
 
@@ -998,60 +998,33 @@ void adapter_set_class_complete(bdaddr_t *bdaddr, uint8_t status)
 		adapter->pending_cod = adapter->wanted_cod;
 }
 
-void adapter_update_tx_power(bdaddr_t *bdaddr, uint8_t status, void *ptr)
+void adapter_update_tx_power(struct btd_adapter *adapter, int8_t tx_power)
 {
-	struct btd_adapter *adapter;
-
-	if (status)
-		return;
-
-	adapter = manager_find_adapter(bdaddr);
-	if (!adapter) {
-		error("Unable to find matching adapter");
-		return;
-	}
-
-	adapter->tx_power = *((int8_t *) ptr);
+	adapter->tx_power = tx_power;
 
 	DBG("inquiry respone tx power level is %d", adapter->tx_power);
 
 	update_ext_inquiry_response(adapter);
 }
 
-void adapter_update_local_name(bdaddr_t *bdaddr, uint8_t status, void *ptr)
+void adapter_update_local_name(struct btd_adapter *adapter, const char *name)
 {
-	read_local_name_rp rp;
-	struct hci_dev *dev;
-	struct btd_adapter *adapter;
-	gchar *name;
+	struct hci_dev *dev = &adapter->dev;
 
-	if (status)
+	if (strncmp(name, dev->name, MAX_NAME_LENGTH) == 0)
 		return;
 
-	adapter = manager_find_adapter(bdaddr);
-	if (!adapter) {
-		error("Unable to find matching adapter");
-		return;
-	}
-
-	dev = &adapter->dev;
-
-	memcpy(&rp, ptr, sizeof(rp));
-	if (strncmp((char *) rp.name, (char *) dev->name, MAX_NAME_LENGTH) == 0)
-		return;
-
-	strncpy((char *) dev->name, (char *) rp.name, MAX_NAME_LENGTH);
+	strncpy(dev->name, name, MAX_NAME_LENGTH);
 
 	if (!adapter->name_stored) {
-		write_local_name(bdaddr, (char *) dev->name);
+		char *name_ptr = dev->name;
 
-		name = g_strdup((char *) dev->name);
+		write_local_name(&adapter->bdaddr, dev->name);
 
 		if (connection)
 			emit_property_changed(connection, adapter->path,
 						ADAPTER_INTERFACE, "Name",
-						DBUS_TYPE_STRING, &name);
-		g_free(name);
+						DBUS_TYPE_STRING, &name_ptr);
 	}
 
 	adapter->name_stored = FALSE;
@@ -1059,46 +1032,26 @@ void adapter_update_local_name(bdaddr_t *bdaddr, uint8_t status, void *ptr)
 	update_ext_inquiry_response(adapter);
 }
 
-void adapter_setname_complete(bdaddr_t *local, uint8_t status)
-{
-	struct btd_adapter *adapter;
-	int err;
-
-	if (status)
-		return;
-
-	adapter = manager_find_adapter(local);
-	if (!adapter) {
-		error("No matching adapter found");
-		return;
-	}
-
-	err = adapter_ops->read_name(adapter->dev_id);
-	if (err < 0)
-		error("Sending getting name command failed: %s (%d)",
-						strerror(errno), errno);
-
-}
-
 static DBusMessage *set_name(DBusConnection *conn, DBusMessage *msg,
 					const char *name, void *data)
 {
 	struct btd_adapter *adapter = data;
 	struct hci_dev *dev = &adapter->dev;
+	char *name_ptr = dev->name;
 
 	if (!g_utf8_validate(name, -1, NULL)) {
 		error("Name change failed: supplied name isn't valid UTF-8");
 		return invalid_args(msg);
 	}
 
-	if (strncmp(name, (char *) dev->name, MAX_NAME_LENGTH) == 0)
+	if (strncmp(name, dev->name, MAX_NAME_LENGTH) == 0)
 		goto done;
 
-	strncpy((char *) adapter->dev.name, name, MAX_NAME_LENGTH);
+	strncpy(dev->name, name, MAX_NAME_LENGTH);
 	write_local_name(&adapter->bdaddr, name);
 	emit_property_changed(connection, adapter->path,
 					ADAPTER_INTERFACE, "Name",
-					DBUS_TYPE_STRING, &name);
+					DBUS_TYPE_STRING, &name_ptr);
 
 	if (adapter->up) {
 		int err = adapter_ops->set_name(adapter->dev_id, name);
@@ -1244,15 +1197,15 @@ sdp_list_t *adapter_get_services(struct btd_adapter *adapter)
 }
 
 struct btd_device *adapter_create_device(DBusConnection *conn,
-						struct btd_adapter *adapter,
-						const char *address)
+					struct btd_adapter *adapter,
+					const char *address, gboolean le)
 {
 	struct btd_device *device;
 	const char *path;
 
 	DBG("%s", address);
 
-	device = device_create(conn, adapter, address);
+	device = device_create(conn, adapter, address, le);
 	if (!device)
 		return NULL;
 
@@ -1311,7 +1264,7 @@ struct btd_device *adapter_get_device(DBusConnection *conn,
 	if (device)
 		return device;
 
-	return adapter_create_device(conn, adapter, address);
+	return adapter_create_device(conn, adapter, address, FALSE);
 }
 
 static gboolean stop_scanning(gpointer user_data)
@@ -1738,7 +1691,10 @@ static DBusMessage *create_device(DBusConnection *conn,
 {
 	struct btd_adapter *adapter = data;
 	struct btd_device *device;
+	struct remote_dev_info *dev, match;
 	const gchar *address;
+	gboolean le;
+	int err;
 
 	if (dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &address,
 						DBUS_TYPE_INVALID) == FALSE)
@@ -1754,11 +1710,20 @@ static DBusMessage *create_device(DBusConnection *conn,
 
 	DBG("%s", address);
 
-	device = adapter_create_device(conn, adapter, address);
+	memset(&match, 0, sizeof(struct remote_dev_info));
+	str2ba(address, &match.bdaddr);
+	match.name_status = NAME_ANY;
+
+	dev = adapter_search_found_devices(adapter, &match);
+	le  = dev ? dev->le : FALSE;
+
+	device = adapter_create_device(conn, adapter, address, le);
 	if (!device)
 		return NULL;
 
-	device_browse(device, conn, msg, NULL, FALSE);
+	err = device_browse(device, conn, msg, NULL, FALSE);
+	if (err < 0)
+		return failed_strerror(msg, -err);
 
 	return NULL;
 }
@@ -1995,94 +1960,14 @@ static GDBusSignalTable adapter_signals[] = {
 	{ }
 };
 
-static inline uint8_t get_inquiry_mode(struct hci_dev *dev)
-{
-	if (dev->features[6] & LMP_EXT_INQ)
-		return 2;
-
-	if (dev->features[3] & LMP_RSSI_INQ)
-		return 1;
-
-	if (dev->manufacturer == 11 &&
-			dev->hci_rev == 0x00 && dev->lmp_subver == 0x0757)
-		return 1;
-
-	if (dev->manufacturer == 15) {
-		if (dev->hci_rev == 0x03 && dev->lmp_subver == 0x6963)
-			return 1;
-		if (dev->hci_rev == 0x09 && dev->lmp_subver == 0x6963)
-			return 1;
-		if (dev->hci_rev == 0x00 && dev->lmp_subver == 0x6965)
-			return 1;
-	}
-
-	if (dev->manufacturer == 31 &&
-			dev->hci_rev == 0x2005 && dev->lmp_subver == 0x1805)
-		return 1;
-
-	return 0;
-}
-
 static int adapter_setup(struct btd_adapter *adapter, const char *mode)
 {
 	struct hci_dev *dev = &adapter->dev;
-	uint8_t events[8] = { 0xff, 0xff, 0xff, 0xff, 0xff, 0x1f, 0x00, 0x00 };
-	uint8_t inqmode;
 	int err;
 	char name[MAX_NAME_LENGTH + 1];
 	uint8_t cls[3];
 
-	if (dev->lmp_ver > 1) {
-		if (dev->features[5] & LMP_SNIFF_SUBR)
-			events[5] |= 0x20;
-
-		if (dev->features[5] & LMP_PAUSE_ENC)
-			events[5] |= 0x80;
-
-		if (dev->features[6] & LMP_EXT_INQ)
-			events[5] |= 0x40;
-
-		if (dev->features[6] & LMP_NFLUSH_PKTS)
-			events[7] |= 0x01;
-
-		if (dev->features[7] & LMP_LSTO)
-			events[6] |= 0x80;
-
-		if (dev->features[6] & LMP_SIMPLE_PAIR) {
-			events[6] |= 0x01;	/* IO Capability Request */
-			events[6] |= 0x02;	/* IO Capability Response */
-			events[6] |= 0x04;	/* User Confirmation Request */
-			events[6] |= 0x08;	/* User Passkey Request */
-			events[6] |= 0x10;	/* Remote OOB Data Request */
-			events[6] |= 0x20;	/* Simple Pairing Complete */
-			events[7] |= 0x04;	/* User Passkey Notification */
-			events[7] |= 0x08;	/* Keypress Notification */
-			events[7] |= 0x10;	/* Remote Host Supported
-						 * Features Notification */
-		}
-
-		if (dev->features[4] & LMP_LE)
-			events[7] |= 0x20;	/* LE Meta-Event */
-
-		adapter_ops->set_event_mask(adapter->dev_id, events,
-							sizeof(events));
-	}
-
-	inqmode = get_inquiry_mode(dev);
-	if (inqmode < 1)
-		return 0;
-
-	err = adapter_ops->write_inq_mode(adapter->dev_id, inqmode);
-	if (err < 0) {
-		error("Can't write inquiry mode for %s: %s (%d)",
-					adapter->path, strerror(-err), -err);
-		return err;
-	}
-
-	if (dev->features[7] & LMP_INQ_TX_PWR)
-		adapter_ops->read_inq_tx_pwr(adapter->dev_id);
-
-	if (dev->features[4] & LMP_LE) {
+	if ((dev->features[4] & LMP_LE) && main_opts.le) {
 		uint8_t simul = (dev->features[6] & LMP_LE_BREDR) ? 0x01 : 0x00;
 		err = adapter_ops->write_le_host(adapter->dev_id, 0x01, simul);
 		if (err < 0) {
@@ -2128,7 +2013,7 @@ static void create_stored_device_from_profiles(char *key, char *value,
 				key, (GCompareFunc) device_address_cmp))
 		return;
 
-	device = device_create(connection, adapter, key);
+	device = device_create(connection, adapter, key, FALSE);
 	if (!device)
 		return;
 
@@ -2151,7 +2036,7 @@ static void create_stored_device_from_linkkeys(char *key, char *value,
 					(GCompareFunc) device_address_cmp))
 		return;
 
-	device = device_create(connection, adapter, key);
+	device = device_create(connection, adapter, key, FALSE);
 	if (device) {
 		device_set_temporary(device, FALSE);
 		adapter->devices = g_slist_append(adapter->devices, device);
@@ -2168,7 +2053,7 @@ static void create_stored_device_from_blocked(char *key, char *value,
 				key, (GCompareFunc) device_address_cmp))
 		return;
 
-	device = device_create(connection, adapter, key);
+	device = device_create(connection, adapter, key, FALSE);
 	if (device) {
 		device_set_temporary(device, FALSE);
 		adapter->devices = g_slist_append(adapter->devices, device);
@@ -2308,8 +2193,8 @@ static void adapter_disable_cod_cache(struct btd_adapter *adapter)
 
 	err = adapter_ops->set_class(adapter->dev_id, adapter->wanted_cod);
 	if (err < 0)
-		error("Adapter class update failed: %s(%d)",
-						strerror(err), err);
+		error("Adapter class update failed: %s (%d)",
+						strerror(-err), -err);
 	else
 		adapter->pending_cod = adapter->wanted_cod;
 }
@@ -2389,7 +2274,6 @@ int adapter_get_discover_type(struct btd_adapter *adapter)
 static int adapter_up(struct btd_adapter *adapter, const char *mode)
 {
 	char srcaddr[18];
-	uint8_t scan_mode;
 	gboolean powered, dev_down = FALSE;
 	int err;
 
@@ -2402,7 +2286,6 @@ static int adapter_up(struct btd_adapter *adapter, const char *mode)
 	adapter->state = STATE_IDLE;
 	adapter->mode = MODE_CONNECTABLE;
 	adapter->cache_enable = TRUE;
-	scan_mode = SCAN_PAGE;
 	powered = TRUE;
 
 	/* Set pairable mode */
@@ -2426,10 +2309,8 @@ static int adapter_up(struct btd_adapter *adapter, const char *mode)
 		write_device_mode(&adapter->bdaddr, onmode);
 
 		return adapter_up(adapter, onmode);
-	} else if (!g_str_equal(mode, "connectable")) {
+	} else if (!g_str_equal(mode, "connectable"))
 		adapter->mode = MODE_DISCOVERABLE;
-		scan_mode = SCAN_PAGE | SCAN_INQUIRY;
-	}
 
 proceed:
 	err = adapter_set_mode(adapter, adapter->mode);
@@ -2438,6 +2319,7 @@ proceed:
 		return err;
 
 	if (adapter->initialized == FALSE) {
+		sdp_init_services_list(&adapter->bdaddr);
 		load_drivers(adapter);
 		clear_blocked(adapter);
 		load_devices(adapter);
@@ -2479,8 +2361,6 @@ int adapter_start(struct btd_adapter *adapter)
 	if (!bacmp(&adapter->bdaddr, BDADDR_ANY)) {
 		int err;
 
-		DBG("Adapter %s without an address", adapter->path);
-
 		err = adapter_ops->read_bdaddr(adapter->dev_id, &adapter->bdaddr);
 		if (err < 0)
 			return err;
@@ -2518,12 +2398,6 @@ int adapter_start(struct btd_adapter *adapter)
 
 	memcpy(dev->features, features, 8);
 
-	if (!(features[6] & LMP_SIMPLE_PAIR))
-		goto setup;
-
-	adapter_ops->init_ssp_mode(adapter->dev_id, &dev->ssp_mode);
-
-setup:
 	adapter_ops->read_link_policy(adapter->dev_id);
 
 	adapter->current_cod = 0;
@@ -2962,10 +2836,27 @@ static void emit_device_found(const char *path, const char *address,
 	g_dbus_send_message(connection, signal);
 }
 
-static char **get_eir_uuids(uint8_t *eir_data, size_t *uuid_count)
+static char **strlist2array(GSList *list)
+{
+	GSList *l;
+	unsigned int i, n;
+	char **array;
+
+	if (list == NULL)
+		return NULL;
+
+	n = g_slist_length(list);
+	array = g_new0(char *, n + 1);
+
+	for (l = list, i = 0; l; l = l->next, i++)
+		array[i] = g_strdup((const gchar *) l->data);
+
+	return array;
+}
+
+static GSList *get_eir_uuids(uint8_t *eir_data, size_t eir_length, GSList *list)
 {
 	uint16_t len = 0;
-	char **uuids;
 	size_t total;
 	size_t uuid16_count = 0;
 	size_t uuid32_count = 0;
@@ -2974,17 +2865,20 @@ static char **get_eir_uuids(uint8_t *eir_data, size_t *uuid_count)
 	uint8_t *uuid32;
 	uint8_t *uuid128;
 	uuid_t service;
+	char *uuid_str;
 	unsigned int i;
 
-	while (len < EIR_DATA_LENGTH - 1) {
-		uint8_t type = eir_data[1];
+	if (eir_data == NULL || eir_length == 0)
+		return list;
+
+	while (len < eir_length - 1) {
 		uint8_t field_len = eir_data[0];
 
 		/* Check for the end of EIR */
 		if (field_len == 0)
 			break;
 
-		switch (type) {
+		switch (eir_data[1]) {
 		case EIR_UUID16_SOME:
 		case EIR_UUID16_ALL:
 			uuid16_count = field_len / 2;
@@ -3007,16 +2901,13 @@ static char **get_eir_uuids(uint8_t *eir_data, size_t *uuid_count)
 	}
 
 	/* Bail out if got incorrect length */
-	if (len > EIR_DATA_LENGTH)
-		return NULL;
+	if (len > eir_length)
+		return list;
 
 	total = uuid16_count + uuid32_count + uuid128_count;
-	*uuid_count = total;
 
 	if (!total)
-		return NULL;
-
-	uuids = g_new0(char *, total + 1);
+		return list;
 
 	/* Generate uuids in SDP format (EIR data is Little Endian) */
 	service.type = SDP_UUID16;
@@ -3025,7 +2916,12 @@ static char **get_eir_uuids(uint8_t *eir_data, size_t *uuid_count)
 
 		val16 = (val16 << 8) + uuid16[0];
 		service.value.uuid16 = val16;
-		uuids[i] = bt_uuid2string(&service);
+		uuid_str = bt_uuid2string(&service);
+		if (g_slist_find_custom(list, uuid_str,
+						(GCompareFunc) strcmp) == NULL)
+			list = g_slist_append(list, uuid_str);
+		else
+			g_free(uuid_str);
 		uuid16 += 2;
 	}
 
@@ -3038,7 +2934,12 @@ static char **get_eir_uuids(uint8_t *eir_data, size_t *uuid_count)
 			val32 = (val32 << 8) + uuid32[k];
 
 		service.value.uuid32 = val32;
-		uuids[i] = bt_uuid2string(&service);
+		uuid_str = bt_uuid2string(&service);
+		if (g_slist_find_custom(list, uuid_str,
+						(GCompareFunc) strcmp) == NULL)
+			list = g_slist_append(list, uuid_str);
+		else
+			g_free(uuid_str);
 		uuid32 += 4;
 	}
 
@@ -3049,15 +2950,21 @@ static char **get_eir_uuids(uint8_t *eir_data, size_t *uuid_count)
 		for (k = 0; k < 16; k++)
 			service.value.uuid128.data[k] = uuid128[16 - k - 1];
 
-		uuids[i] = bt_uuid2string(&service);
+		uuid_str = bt_uuid2string(&service);
+		if (g_slist_find_custom(list, uuid_str,
+						(GCompareFunc) strcmp) == NULL)
+			list = g_slist_append(list, uuid_str);
+		else
+			g_free(uuid_str);
 		uuid128 += 16;
 	}
 
-	return uuids;
+	return list;
 }
 
 void adapter_emit_device_found(struct btd_adapter *adapter,
-				struct remote_dev_info *dev, uint8_t *eir_data)
+				struct remote_dev_info *dev,
+				uint8_t *eir_data, size_t eir_length)
 {
 	struct btd_device *device;
 	char peer_addr[18], local_addr[18];
@@ -3065,8 +2972,7 @@ void adapter_emit_device_found(struct btd_adapter *adapter,
 	dbus_bool_t paired = FALSE;
 	dbus_int16_t rssi = dev->rssi;
 	char *alias;
-	char **uuids = NULL;
-	size_t uuid_count = 0;
+	size_t uuid_count;
 
 	ba2str(&dev->bdaddr, peer_addr);
 	ba2str(&adapter->bdaddr, local_addr);
@@ -3075,6 +2981,26 @@ void adapter_emit_device_found(struct btd_adapter *adapter,
 	if (device)
 		paired = device_is_paired(device);
 
+	/* Extract UUIDs from extended inquiry response if any */
+	dev->services = get_eir_uuids(eir_data, eir_length, dev->services);
+	uuid_count = g_slist_length(dev->services);
+	if (dev->services && dev->uuid_count != uuid_count) {
+		g_strfreev(dev->uuids);
+		dev->uuids = strlist2array(dev->services);
+		dev->uuid_count = uuid_count;
+	}
+
+	if (dev->le) {
+		emit_device_found(adapter->path, paddr,
+				"Address", DBUS_TYPE_STRING, &paddr,
+				"RSSI", DBUS_TYPE_INT16, &rssi,
+				"Name", DBUS_TYPE_STRING, &dev->name,
+				"Paired", DBUS_TYPE_BOOLEAN, &paired,
+				"UUIDs", DBUS_TYPE_ARRAY, &dev->uuids,
+				dev->uuid_count, NULL);
+		return;
+	}
+
 	icon = class_to_icon(dev->class);
 
 	if (!dev->alias) {
@@ -3086,10 +3012,6 @@ void adapter_emit_device_found(struct btd_adapter *adapter,
 	} else
 		alias = g_strdup(dev->alias);
 
-	/* Extract UUIDs from extended inquiry response if any*/
-	if (eir_data != NULL)
-		uuids = get_eir_uuids(eir_data, &uuid_count);
-
 	emit_device_found(adapter->path, paddr,
 			"Address", DBUS_TYPE_STRING, &paddr,
 			"Class", DBUS_TYPE_UINT32, &dev->class,
@@ -3099,17 +3021,15 @@ void adapter_emit_device_found(struct btd_adapter *adapter,
 			"Alias", DBUS_TYPE_STRING, &alias,
 			"LegacyPairing", DBUS_TYPE_BOOLEAN, &dev->legacy,
 			"Paired", DBUS_TYPE_BOOLEAN, &paired,
-			"UUIDs", DBUS_TYPE_ARRAY, &uuids, uuid_count,
+			"UUIDs", DBUS_TYPE_ARRAY, &dev->uuids, dev->uuid_count,
 			NULL);
 
 	g_free(alias);
-	g_strfreev(uuids);
 }
 
-void adapter_update_found_devices(struct btd_adapter *adapter, bdaddr_t *bdaddr,
-				int8_t rssi, uint32_t class, const char *name,
-				const char *alias, gboolean legacy,
-				name_status_t name_status, uint8_t *eir_data)
+static struct remote_dev_info *get_found_dev(struct btd_adapter *adapter,
+						const bdaddr_t *bdaddr,
+						gboolean *new_dev)
 {
 	struct remote_dev_info *dev, match;
 
@@ -3119,36 +3039,103 @@ void adapter_update_found_devices(struct btd_adapter *adapter, bdaddr_t *bdaddr,
 
 	dev = adapter_search_found_devices(adapter, &match);
 	if (dev) {
+		*new_dev = FALSE;
 		/* Out of range list update */
 		adapter->oor_devices = g_slist_remove(adapter->oor_devices,
 							dev);
-
-		if (rssi == dev->rssi)
-			return;
-
-		goto done;
+	} else {
+		*new_dev = TRUE;
+		dev = g_new0(struct remote_dev_info, 1);
+		bacpy(&dev->bdaddr, bdaddr);
+		adapter->found_devices = g_slist_prepend(adapter->found_devices,
+									dev);
 	}
 
-	dev = g_new0(struct remote_dev_info, 1);
+	return dev;
+}
 
-	bacpy(&dev->bdaddr, bdaddr);
-	dev->class = class;
-	if (name)
-		dev->name = g_strdup(name);
-	if (alias)
-		dev->alias = g_strdup(alias);
-	dev->legacy = legacy;
-	dev->name_status = name_status;
+static uint8_t extract_eir_flags(uint8_t *eir_data)
+{
+	if (eir_data[0] == 0)
+		return 0;
 
-	adapter->found_devices = g_slist_prepend(adapter->found_devices, dev);
+	if (eir_data[1] != EIR_FLAGS)
+		return 0;
+
+	/* For now, only one octet is used for flags */
+	return eir_data[2];
+}
+
+void adapter_update_device_from_info(struct btd_adapter *adapter,
+						le_advertising_info *info)
+{
+	struct remote_dev_info *dev;
+	bdaddr_t bdaddr;
+	gboolean new_dev;
+	int8_t rssi;
+
+	rssi = *(info->data + info->length);
+	bdaddr = info->bdaddr;
+
+	dev = get_found_dev(adapter, &bdaddr, &new_dev);
+
+	if (new_dev) {
+		dev->le = TRUE;
+		dev->evt_type = info->evt_type;
+	} else if (dev->rssi == rssi)
+		return;
 
-done:
 	dev->rssi = rssi;
 
 	adapter->found_devices = g_slist_sort(adapter->found_devices,
 						(GCompareFunc) dev_rssi_cmp);
 
-	adapter_emit_device_found(adapter, dev, eir_data);
+	if (info->length) {
+		uint8_t type;
+		char *tmp_name = bt_extract_eir_name(info->data, &type);
+		if (tmp_name) {
+			g_free(dev->name);
+			dev->name = tmp_name;
+		}
+
+		dev->flags = extract_eir_flags(info->data);
+	}
+
+	/* FIXME: check if other information was changed before emitting the
+	 * signal */
+	adapter_emit_device_found(adapter, dev, info->data, info->length);
+}
+
+void adapter_update_found_devices(struct btd_adapter *adapter, bdaddr_t *bdaddr,
+				int8_t rssi, uint32_t class, const char *name,
+				const char *alias, gboolean legacy,
+				name_status_t name_status, uint8_t *eir_data)
+{
+	struct remote_dev_info *dev;
+	gboolean new_dev;
+
+	dev = get_found_dev(adapter, bdaddr, &new_dev);
+
+	if (new_dev) {
+		if (name)
+			dev->name = g_strdup(name);
+
+		if (alias)
+			dev->alias = g_strdup(alias);
+
+		dev->le = FALSE;
+		dev->class = class;
+		dev->legacy = legacy;
+		dev->name_status = name_status;
+	} else if (dev->rssi == rssi)
+		return;
+
+	dev->rssi = rssi;
+
+	adapter->found_devices = g_slist_sort(adapter->found_devices,
+						(GCompareFunc) dev_rssi_cmp);
+
+	adapter_emit_device_found(adapter, dev, eir_data, EIR_DATA_LENGTH);
 }
 
 int adapter_remove_found_device(struct btd_adapter *adapter, bdaddr_t *bdaddr)
@@ -3668,10 +3655,9 @@ int btd_adapter_remove_bonding(struct btd_adapter *adapter, bdaddr_t *bdaddr)
 }
 
 int btd_adapter_request_authentication(struct btd_adapter *adapter,
-					uint16_t handle, uint8_t *status)
+							uint16_t handle)
 {
-	return adapter_ops->request_authentication(adapter->dev_id,
-							handle, status);
+	return adapter_ops->request_authentication(adapter->dev_id, handle);
 }
 
 int btd_adapter_pincode_reply(struct btd_adapter *adapter, bdaddr_t *bdaddr,
@@ -3703,11 +3689,6 @@ int btd_adapter_read_scan_enable(struct btd_adapter *adapter)
 	return adapter_ops->read_scan_enable(adapter->dev_id);
 }
 
-int btd_adapter_read_ssp_mode(struct btd_adapter *adapter)
-{
-	return adapter_ops->read_ssp_mode(adapter->dev_id);
-}
-
 int btd_adapter_read_local_ext_features(struct btd_adapter *adapter)
 {
 	return adapter_ops->read_local_ext_features(adapter->dev_id);
